//
//  MWDocument.mm
//  MySQL Workbench
//
//  Created by Alfredo Kojima on 05/6/17.
//  Copyright MySQL AB 2005 . All rights reserved.
//

#import "MWDocument.h"
#import "MWCanvasViewTab.h"
#import "MWLayerTree.h"

#import <WBCanvas/MCanvasScrollView.h>
#import <WBCanvas/MGenericCanvasView.h>
#import <MySQLToolsCommon/MTabView.h>
#import <MySQLToolsCommon/MPreferences.h>
#import <MySQLToolsCommon/mxUtils.h>
#import <MySQLGRT/MGRT.h>
#import <MySQLGRT/MGRTModuleLoader.h>
#import <MySQLGRT/MGRTTableEditor.h>
#import <MySQLGRT/MGRTViewEditor.h>
#import <MySQLGRT/MGRTRoutineGroupEditor.h>
#import <MySQLGRT/MGRTSchemaEditor.h>

#include <WBCanvas/myx_gc_canvas.h>
#include <WBCanvas/myx_gc_texture.h>

#include <WBCanvas/myx_grt_wb_public_interface.h>

#import "MWReverseEngineering.h"


@interface MSolidColorImageRep : NSImageRep
{
  NSColor *_color;
}
- (id)initWithColor:(NSColor*)color;
@end


@implementation MSolidColorImageRep
- (id)initWithColor:(NSColor*)color
{
  self= [super init];
  if (self)
    _color= [color retain];
  return self;
}

- (void)dealloc
{
  [_color release];
  [super dealloc];
}

- (BOOL)draw
{
  NSRect rect;
  rect.origin= NSMakePoint(0, 0);
  rect.size= [self size];
  [_color set];
  NSRectFill(rect);
  [[NSColor lightGrayColor] set];
  [NSBezierPath strokeRect:rect];
  return YES;
}
@end



static const NSSize MWDefaultCanvasSize= {10000.0, 7000.0};

@implementation MWDocument


- (void)viewAdded:(id)aView
{
  CGCView *view= (CGCView*)[aView pointerValue];
  string name= view->property("name",0);
  
  [_canvas createCanvasViewWithView:view];
  
  MWCanvasViewTab *cview= [[MWCanvasViewTab alloc] initWithContent:_scrollView
                                                            canvas:_canvas
                                                              view:view
                                                              name:[NSString stringWithUTF8String:name.c_str()]];
  
  [tabView addTabViewItem:cview];
  
  [cview release];
}


- (void)refreshWhenIdle:(id)dummy
{
  if ([[[NSThread currentThread] threadDictionary] objectForKey:@"mainThread"])
  {
    if (!_postedIdle)
    {
      _postedIdle= YES;
      
      NSNotification* notification = [NSNotification notificationWithName:@"MWBIdleNotification" object:nil];
      [[NSNotificationQueue defaultQueue] enqueueNotification:notification postingStyle:NSPostWhenIdle];
    }
  }
  else
    [self performSelectorOnMainThread:@selector(refreshWhenIdle:)
                           withObject:nil
                        waitUntilDone:NO];
}


- (void)switchToView:(id)viewValue
{
  CGCView* view= (CGCView*)[viewValue pointerValue];
  
  for (int i= 0; i < [tabView numberOfPages]; i++)
  {
    MWCanvasViewTab *tab= [tabView pageAtIndex:i];
    
    if ([tab gcView] == view)
    {
      [tabView selectPage:i];
      break;
    }
  }
}


static void viewCallback(MYX_GRT *grt, CGCView *view, MYX_WB_GC_CHANGE change, void *data)
{
  MWDocument *self= (MWDocument*)data;
  
  switch (change)
  {
    case MYX_WBGC_ADDED:
      [self performSelectorOnMainThread:@selector(viewAdded:)
                             withObject:[NSValue valueWithPointer:view]
                          waitUntilDone:NO];
      break;
    case MYX_WBGC_SWITCHED:
      [self performSelectorOnMainThread:@selector(switchToView:)
                             withObject:[NSValue valueWithPointer:view]
                          waitUntilDone:NO];
      break;
    case MYX_WBGC_RUBBERBAND_STARTED:
    case MYX_WBGC_RUBBERBAND_STOPPED:
      [self refreshWhenIdle:nil];
      break;
    default:
      break;
  }
  [self updateChangeCount: NSChangeDone];
}


static void layerCallback(MYX_GRT *grt, MYX_GRT_VALUE *layer, MYX_WB_GC_CHANGE change, void *data)
{
  MWDocument *self= (MWDocument*)data;
  id newNode= nil;

  switch (change)
  {
    case MYX_WBGC_ADDED:
      newNode= [self->_layerDS handleLayerAdd:layer];
      break;
    case MYX_WBGC_MODIFIED:
      [self->_layerDS handleLayerChange:layer];
      break;
    case MYX_WBGC_REMOVED:
      [self->_layerDS handleLayerDelete:layer];
      break;
  }
  [self->layerTree reloadData];
  if (newNode)
    [self->layerTree expandItem:newNode];
  
  [self refreshWhenIdle:nil];
  
  [self updateChangeCount: NSChangeDone];
}


static void elementCallback(MYX_GRT *grt, MYX_GRT_VALUE *element, MYX_WB_GC_CHANGE change, void *data)
{
  MWDocument *self= (MWDocument*)data;

  switch (change)
  {
    case MYX_WBGC_ADDED:
      [self->_layerDS handleElementAdd:element];
      [self->layerTree reloadData];
      break;
    case MYX_WBGC_REMOVED:
      [self->_layerDS handleElementDelete:element];
      [self->layerTree reloadData];
      break;
    case MYX_WBGC_MODIFIED:
      if (element)
      {
        [self->_layerDS handleElementChange:element];
        [self->layerTree reloadData];
      }
      break;
  }
  
  [self->schemaTree reloadData];
  
  [self refreshWhenIdle:nil];
  
  [self updateChangeCount: NSChangeDone];
}


// for obj-c module support
- (BOOL)exportsMethod:(NSString*)method
{
  if ([method isEqualTo:@"editSchema:"])
    return YES;
  if ([method isEqualTo:@"editTable:"])
    return YES;
  if ([method isEqualTo:@"editView:"])
    return YES;
  if ([method isEqualTo:@"editRoutineGroup:"])
    return YES;
  return NO;
}


- (void)toggleDrawer:(id)sender
{
  if ([drawer state] == NSDrawerOpenState)
  {
    [drawer close];
    [sender setTitle:@"Open Drawer"];
  }
  else
  {
    [drawer open];
    [sender setTitle:@"Close Drawer"];
  }
}


- (IBAction)createSchema:(id)sender
{
  MGRTSchemaEditor *editor= [[MGRTSchemaEditor alloc] init];
  
  [editor setDelegate:self];
  [editor setMGRT:_grt catalogs:MGRTValue([_grt globalValue:"/workbench/catalog"])];
  [editor createNew];
  [editor showWindow:nil];
  
  [_openEditors setObject:editor forKey:NSStr([editor objectId])];
  [editor release];
}


- (void)editSelectedSchemaObject:(id)sender
{
  MYX_GRT_VALUE *object= [catalogDS objectAtNode:[schemaTree itemAtRow:[schemaTree selectedRow]]];

  if (myx_grt_value_get_type(object) == MYX_DICT_VALUE)
  {
    [_grt performModule:@"Base"
              procedure:@"editObj"
              arguments:[NSArray arrayWithObject:[NSValue valueWithPointer:object]]];
  }
}


- (MYX_GRT_VALUE*)editSchema:(NSArray*)args
{
  MYX_GRT_VALUE *object= (MYX_GRT_VALUE*)[[args objectAtIndex:0] pointerValue];
  
  if (object)
  {
    NSString *schemaId= [NSString stringWithUTF8String:myx_grt_dict_id_item_as_string(object)];
    MGRTSchemaEditor *editor;
    
    if ((editor= [_openEditors objectForKey:schemaId]))
    {
      [[editor window] makeKeyAndOrderFront:nil];
    }
    else
    {
      editor= [[MGRTSchemaEditor alloc] init];
      [editor setDelegate:self];
      [editor setMGRT:_grt catalogs:MGRTValue([_grt globalValue:"/workbench/catalog"])];
      [editor editObject:object];
      [editor showWindow:nil];
      
      [_openEditors setObject:editor forKey:schemaId];
      [editor release];
    }
  }
  return NULL;
}


- (MYX_GRT_VALUE*)editTable:(NSArray*)args
{
  MYX_GRT_VALUE *object= (MYX_GRT_VALUE*)[[args objectAtIndex:0] pointerValue];

  if (object)
  {
    NSString *tableId= [NSString stringWithUTF8String:myx_grt_dict_id_item_as_string(object)];
    MGRTTableEditor *editor;
    
    if ((editor= [_openEditors objectForKey:tableId]))
    {
      [[editor window] makeKeyAndOrderFront:nil];
    }
    else
    {
      editor= [[MGRTTableEditor alloc] init];
      [editor setDelegate:self];
      [editor setMGRT:_grt catalogs:MGRTValue([_grt globalValue:"/workbench/catalog"])];
      [editor editObject:object];
      [editor showWindow:nil];
      
      [_openEditors setObject:editor forKey:tableId];
      [editor release];
    }
  }
  return NULL;
}


- (MYX_GRT_VALUE*)editView:(NSArray*)args
{
  MYX_GRT_VALUE *object= (MYX_GRT_VALUE*)[[args objectAtIndex:0] pointerValue];
  
  if (object)
  {
    NSString *viewId= [NSString stringWithUTF8String:myx_grt_dict_id_item_as_string(object)];
    MGRTViewEditor *editor;
    
    if ((editor= [_openEditors objectForKey:viewId]))
    {
      [[editor window] makeKeyAndOrderFront:nil];
    }
    else
    {
      editor= [[MGRTViewEditor alloc] init];
      [editor setDelegate:self];
      [editor setMGRT:_grt catalogs:MGRTValue([_grt globalValue:"/workbench/catalog"])];
      [editor editObject:object];
      [editor showWindow:nil];
      
      [_openEditors setObject:editor forKey:viewId];
      [editor release];
    }
  }
  return NULL;
}


- (MYX_GRT_VALUE*)editRoutineGroup:(NSArray*)args
{
  MYX_GRT_VALUE *object= (MYX_GRT_VALUE*)[[args objectAtIndex:0] pointerValue];
  
  if (object)
  {
    NSString *viewId= [NSString stringWithUTF8String:myx_grt_dict_id_item_as_string(object)];
    MGRTViewEditor *editor;
    
    if ((editor= [_openEditors objectForKey:viewId]))
    {
      [[editor window] makeKeyAndOrderFront:nil];
    }
    else
    {
      editor= [[MGRTRoutineGroupEditor alloc] init];
      [editor setDelegate:self];
      [editor setMGRT:_grt catalogs:MGRTValue([_grt globalValue:"/workbench/catalog"])];
      [editor editObject:object];
      [editor showWindow:nil];
      
      [_openEditors setObject:editor forKey:viewId];
      [editor release];
    }
  }
  return NULL;
}


- (id)init
{
  self = [super init];
  if (self) 
  {    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(performIdleTasks:)
                                                 name:@"MWBIdleNotification"
                                               object:nil];
    
    _openEditors= [[NSMutableDictionary alloc] init];
    
    _grt= [[MGRT alloc] init];
    
    _grtShell= [[MGRTShell alloc] initWithMGRT:_grt];
    [_grt initializeGRT:[[NSBundle bundleForClass:[MGRT class]] resourcePath]];
    [_grt scanStructsInPath:[[NSBundle mainBundle] resourcePath]];
    myx_register_grt_wb_module([_grt grt]);

    // load /app stuff
    if (![_grt loadAppDictFromPath:[[MPreferences preferences] pathForFile:@"Workbench/workbench_app_data.xml"]])
      [_grt globalAppDict]; // this will create /app in case it doesnt exist

    [_grt scanModulesInPath:[[NSBundle mainBundle] resourcePath]];

    [[_grt objCLoader] registerObject:self asModule:@"WorkbenchUi"];

    [_grt performModule:@"Workbench"
              procedure:@"registerEditors"
              arguments:nil];

    _toolCursors= [[NSArray arrayWithObjects:
      [NSCursor arrowCursor],
      [NSCursor openHandCursor],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"rubber"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"table"] hotSpot:NSMakePoint(4,4)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"view"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"routine"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"res11"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"res1n"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"resnm"] hotSpot:NSMakePoint(0,0)] autorelease],      
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"image"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"note"] hotSpot:NSMakePoint(0,0)] autorelease],
      [[[NSCursor alloc] initWithImage:[NSImage imageNamed:@"layer"] hotSpot:NSMakePoint(0,0)] autorelease],
      nil] retain];
    
//    _layerIcon= [[NSImage imageNamed:@""] retain];
  }
  return self;
}

- (void)dealloc
{
  [_grt performModule:@"Workbench"
            procedure:@"shutdownWorkbench"
            arguments:nil];
  
  [[NSNotificationCenter defaultCenter] removeObserver:self];
  [_layerIcon release];
  [_tableIcon release];
  [_toolCursors release];
  [_canvas release];
  [_grtShell hide:nil];
  [_grtShell release];
  [_grt release];
  [_scrollView release];
  [super dealloc];
}

- (NSString *)windowNibName
{
  return @"DocumentWindow";
}


- (void)performIdleTasks:(NSNotification*)notif
{
  [[_canvas openGLContext] makeCurrentContext];
  myx_grt_wb_bridge_process_pending([_grt grt]);
  _postedIdle= NO;
  [_canvas setNeedsDisplay:YES];
}


- (void)setupGRT
{
  MYX_GRT_VALUE *result;
  g_log_set_always_fatal(G_LOG_LEVEL_CRITICAL);
  
  if (!(result= [_grt performModule:@"Workbench"
                           function:@"initWorkbench"
                          arguments:[NSArray arrayWithObject:[[NSBundle mainBundle] resourcePath]]]))
  {
    NSRunAlertPanel(@"Error",@"The GRT environment for the Workbench could not be initialized.\nPlease verify your installation.",
                    @"OK",nil,nil);
  }
  else
    myx_grt_value_release(result);

  myx_grt_wb_bridge_initialize([_grt grt], [_canvas canvas], "/workbench/model");

  myx_grt_wb_bridge_set_callbacks([_grt grt], self,
                                  viewCallback,
                                  layerCallback,
                                  elementCallback);
}


- (void)showGRTShell:(id)sender
{
  [_grtShell show:nil];
}


- (void)setupMenu:(NSMenu*)menu withColors:(NSArray*)colors
{
  int i;
  
  while ([menu numberOfItems])
    [menu removeItemAtIndex:0];
  
  for (i= 0; i < [colors count]; i+=2)
  {
    id item;
    unsigned int color;
    NSImage *image= [[NSImage alloc] initWithSize:NSMakeSize(20,15)];
    NSImageRep *imageRep;
    NSScanner *scanner= [NSScanner scannerWithString:[colors objectAtIndex:i+1]];
    [scanner setScanLocation:1];
    [scanner scanHexInt:&color];
    imageRep= [[MSolidColorImageRep alloc] initWithColor:[NSColor colorWithDeviceRed:((color>>16)&0xff)/255.0
                                                                               green:((color>>8)&0xff)/255.0
                                                                                blue:(color&0xff)/255.0
                                                                               alpha:1.0]];
    [image addRepresentation:imageRep];
    [imageRep release];
        
    item= [[NSMenuItem alloc] init];
    [item setTitle:@""];
    [item setImage:image];
    [menu addItem:item];
    [item setRepresentedObject:[colors objectAtIndex:i]];
    [image release];
    [item release];
  }
}


- (void)setupMenu:(NSMenu*)menu 
        withItems:(MYX_GRT_VALUE*)items
         titleKey:(const char*)key
{
  int i, c= myx_grt_list_item_count(items);
  
  while ([menu numberOfItems])
    [menu removeItemAtIndex:0];
  
  for (i= 0; i < c; i++)
  {
    MYX_GRT_VALUE *item= myx_grt_list_item_get(items, i);
    
    [menu addItemWithTitle:key ? [NSString stringWithUTF8String:myx_grt_dict_item_get_as_string(item, key)] : [NSString stringWithUTF8String:myx_grt_value_as_string(item)]
                    action:nil
             keyEquivalent:@""];
    
    [[menu itemAtIndex:i] setRepresentedObject:[NSValue valueWithPointer:item]];
  }
}


- (void)setupMenu:(NSMenu*)menu 
 withCharsetItems:(MYX_GRT_VALUE*)items
{
  int i, c= myx_grt_list_item_count(items);
  
  while ([menu numberOfItems])
    [menu removeItemAtIndex:0];
  
  for (i= 0; i < c; i++)
  {
    MYX_GRT_VALUE *item= myx_grt_list_item_get(items, i);
    NSString *name= [NSString stringWithUTF8String:myx_grt_dict_item_get_as_string(item, "name")];
    MYX_GRT_VALUE *collations= myx_grt_dict_item_get_value(item, "collations");
    int j, jc;
    
    jc= myx_grt_list_item_count(collations);
    for (j= 0; j < jc; j++)
    {
      const char *coll= myx_grt_value_as_string(myx_grt_list_item_get(collations, j));
      
      [menu addItemWithTitle:NSStr(coll)
                      action:nil
               keyEquivalent:@""];
      
      [[menu itemAtIndex:[menu numberOfItems]-1] setRepresentedObject:name];
    }
  }
}


- (void)windowControllerDidLoadNib:(NSWindowController *)aController
{
  CGenericCanvas *canvas;
  NSBundle *bundle= [NSBundle mainBundle];
  NSRect rect;
  NSArray *zoomSteps;
  unsigned int i;

  [super windowControllerDidLoadNib:aController];

  [tabView setMaxTabSize:100];
  [tabView setDelegate:self];
  
  // setup canvas
  rect.origin= NSMakePoint(0, 0);
  rect.size= [tabView contentSize];
  _scrollView= [[MCanvasScrollView alloc] initWithFrame:rect];
  rect.size= [_scrollView contentSize];
  
  _canvas= [[MWEditCanvasView alloc] initWithFrame:rect];
  [_canvas setAutoresizingMask:NSViewWidthSizable|NSViewHeightSizable];
  [_scrollView setContentCanvas:_canvas];
  [_canvas setNextResponder:[modelWindow contentView]];
  [_canvas setMGRT:_grt];
  [_canvas setStatusText:statusText];
  [_canvas registerForDraggedTypes:[NSArray arrayWithObject:MGRTTreeNodePboardType]];

  
  myx_grt_wb_load_style_file([_canvas canvas],
                             [[bundle pathForResource:@"layout.styles.db.connection"
                                               ofType:@"xml"] UTF8String],
                             NULL);
  myx_grt_wb_load_style_file([_canvas canvas],
                             [[bundle pathForResource:@"layout.styles.db.routine"
                                               ofType:@"xml"] UTF8String],
                             NULL);
  myx_grt_wb_load_style_file([_canvas canvas],
                             [[bundle pathForResource:@"layout.styles.db.table"
                                               ofType:@"xml"] UTF8String],
                             NULL);
  myx_grt_wb_load_style_file([_canvas canvas],
                             [[bundle pathForResource:@"layout.styles.db.view"
                                               ofType:@"xml"] UTF8String],
                             NULL);
  myx_grt_wb_load_style_file([_canvas canvas],
                             [[bundle pathForResource:@"layout.styles.db"
                                               ofType:@"xml"] UTF8String],
                             NULL);
  {
    MYX_GRT_VALUE *presets= [_grt performModule:@"Workbench"
                                       function:@"getColorPresets"
                                      arguments:nil];
    if (!presets)
      NSLog(@"Error getting color presets.");
    else
    {
      NSMutableArray *colors= [[NSMutableArray alloc] init];
      MYX_GRT_VALUE *colorList;
      int i, j;
      struct {
        const char *name;
        NSString *path;
        NSMenu *menu;
      } colorPresetInfo[] = 
      {
      {"table", @"layout.styles.db.table.tmpl", [[tableOptions viewWithTag:1] menu]},
      {"view", @"layout.styles.db.view.tmpl", [[viewOptions viewWithTag:1] menu]},
      {"layer", @"layout.styles.layer.tmpl", [[layerOptions viewWithTag:1] menu]},
      {"note", @"layout.styles.note.tmpl", [[noteOptions viewWithTag:1] menu]},
//      {"package", @"layout.styles.package.tmpl", [[routineOptions viewWithTag:1] menu]},
      {NULL, NULL, nil}
      };

      for (j= 0; colorPresetInfo[j].name; j++)
      {
        [colors removeAllObjects];
        colorList= myx_grt_bridge_dict_item_get_value(presets, colorPresetInfo[j].name, 0);
        myx_grt_wb_load_style_file([_canvas canvas],
                                   [[bundle pathForResource:colorPresetInfo[j].path
                                                     ofType:@"xml"] UTF8String],
                                   colorList);
        for (i= 0; i < myx_grt_bridge_list_item_count(colorList); i++)
        {
          MYX_GRT_VALUE *item= myx_grt_bridge_list_item_get(colorList, i, 0);
          [colors addObject:[NSString stringWithUTF8String:myx_grt_bridge_value_as_string(myx_grt_bridge_dict_item_get_value(item, "colorName", 0))]];
          [colors addObject:[NSString stringWithUTF8String:myx_grt_bridge_value_as_string(myx_grt_bridge_dict_item_get_value(item, "color", 0))]];
        }      
        [self setupMenu:colorPresetInfo[j].menu
             withColors:colors];
      }
      [colors release];
      myx_grt_value_release(presets);
    }
  }
  
  if (![_canvas loadLayoutsFromFile:[bundle pathForResource:@"layout.figures.db"
                                                     ofType:@"xml"]])
    NSLog(@"Error reading figures file.");

  canvas= [_canvas canvas];

  [_canvas setDelegate:self];
  [_canvas setBaseSize:MWDefaultCanvasSize];
  [_scrollView setAutoresizingMask:NSViewWidthSizable|NSViewHeightSizable];

  NSNib *nib= [[NSNib alloc] initWithNibNamed:@"ScrollTools" bundle:[NSBundle mainBundle]];
  [nib instantiateNibWithOwner:self
               topLevelObjects:nil];
  [nib release];
  
  [zoomPop removeAllItems];
  zoomSteps= [_canvas getZoomSteps];
  for (i= 0; i < [zoomSteps count]; i++)
  {
    [zoomPop addItemWithTitle:[zoomSteps objectAtIndex:i]];
  }
  
  [_scrollView setHAccessory:scrollTools];
  
  [self setupGRT];

  _layerDS= [[MWLayerDataSource alloc] initWithGRT:[_grt grt]];
  [layerTree setDataSource:_layerDS];
  [layerTree setDelegate:self];
//  [_layerDS setRootLayerPath:@"/workbench/model/views/0/layers"];
  [layerTree reloadData];
  
  if (!_loadedDocumentData)
  {
    MYX_GRT_VALUE *result;
    if (!(result= [_grt performModule:@"Workbench"
                             function:@"newDocument"
                            arguments:[NSArray arrayWithObject:@"Mysql"]]))
    {
      
    }
    else
      myx_grt_value_release(result);
  }
  else
  {
    MYX_GRT_VALUE *result;
    
    result= [_grt performModule:@"Workbench"
                       function:@"loadDocumentData"
                      arguments:[NSArray arrayWithObject:[NSValue valueWithPointer:_loadedDocumentData]]];
    if (result)
    {
      myx_grt_value_release(result);
    }
    
    [statusText setStringValue:[NSString stringWithFormat:@"Loaded Workbench model '%@'", [self displayName]]];
    
    myx_grt_value_release(_loadedDocumentData);
    _loadedDocumentData= 0;    
  }

  [catalogDS setMGRT:_grt];
  [catalogDS setRootValue:[_grt globalValue:"/workbench/catalog/schemata"]];
  [catalogDS setDisplayInfo:[NSDictionary dictionaryWithObjectsAndKeys:
    [NSArray arrayWithObjects:@"tables",@"views",@"routineGroups",@"routines",nil], @"db.mysql.Schema",
    nil]];
  [catalogDS setIconInfo:[NSDictionary dictionaryWithObjectsAndKeys:
    [NSImage imageNamed:@"db.Schema.16x16"], @"db.mysql.Schema",
    [NSImage imageNamed:@"db.Table.16x16"], @"db.mysql.Table",
    [NSImage imageNamed:@"db.View.16x16"], @"db.mysql.View",
    [NSImage imageNamed:@"db.Routine.16x16"], @"db.mysql.Routine",
    nil]];
  [schemaTree registerForDraggedTypes:[NSArray arrayWithObject:MGRTTreeNodePboardType]];
  [schemaTree reloadData];
  [schemaTree setDoubleAction:@selector(editSelectedSchemaObject:)];
  [schemaTree setTarget:self];

  [self updateChangeCount:NSChangeCleared];
  
  [self setupMenu:[[tableOptions viewWithTag:2] menu]
        withItems:[_grt globalValue:"/workbench/catalog/schemata"]
         titleKey:"name"];
  
  [self setupMenu:[[viewOptions viewWithTag:2] menu]
        withItems:[_grt globalValue:"/workbench/catalog/schemata"]
         titleKey:"name"];

  [self setupMenu:[[routineOptions viewWithTag:2] menu]
        withItems:[_grt globalValue:"/workbench/catalog/schemata"]
         titleKey:"name"];

  
  [self setupMenu:[[tableOptions viewWithTag:4] menu]
 withCharsetItems:[_grt globalValue:"/rdbmsMgmt/rdbms/Mysql/characterSets"]];

  
  [[tableOptions viewWithTag:4] selectItemWithTitle:@"utf8_general_ci"];

  [self updateToolOptions];
//  [self toggleToolHelp:YES];
  
  [drawer open];
}


- (BOOL)readFromFile:(NSString *)fileName ofType:(NSString *)docType
{
  _loadedDocumentData= myx_grt_retrieve_from_file([_grt grt],
                                                  [fileName UTF8String]);
  if (_loadedDocumentData != 0)
    return YES;
  else
    return NO;
}


- (BOOL)writeToFile:(NSString *)fileName ofType:(NSString *)docType
{
  MYX_GRT_ERROR error;
  
  error= myx_grt_store_to_file([_grt grt],
                               [_grt globalValue:"/workbench"],
                               [fileName UTF8String]);
  if (error != MYX_GRT_NO_ERROR)
  {
    NSRunAlertPanel(@"Error Saving Model",
                    [MGRT errorText:error],
                    @"OK",nil,nil);
    return NO;
  }
  return YES;
}


- (BOOL)writeToURL:(NSURL *)absoluteURL ofType:(NSString *)typeName error:(NSError **)outError
{
  return [self writeToFile:[absoluteURL path] ofType:typeName];
}


- (BOOL)readFromURL:(NSURL *)absoluteURL ofType:(NSString *)typeName error:(NSError **)outError
{
  if ([absoluteURL isFileURL])
  {
    return [self readFromFile:[absoluteURL path] ofType:typeName];
  }
  else
    return NO;
}




- (IBAction)reverseEngineer:(id)sender
{
  MWReverseEngineering *reveng= [[MWReverseEngineering alloc] initWithMGRT:_grt];

  [reveng run];
  
  [_grt saveAppDict];
}


- (IBAction)toggleOverview:(id)sender
{
  [_canvas toggleOverview:nil];
}


- (void)generateScript:(id)sender
{
  NSSavePanel *panel= [NSSavePanel savePanel];
  
  [panel setTitle:@"Generate Database Create SQL Script"];
  [panel setRequiredFileType:@"sql"];
  
  if ([panel runModal] == NSFileHandlingPanelOKButton)
  {
    NSString *data= [_grt performModule:@"Workbench"
                         stringFunction:@"generateSqlCreateScript"
                              arguments:nil];
    
    [data writeToFile:[panel filename] atomically:NO];
    
    [statusText setStringValue:[NSString stringWithFormat:@"Wrote SQL script to %@", [panel filename]]];
  }
}


- (void)tabViewChangedPage:(MTabView*)aTabView
{
  MWCanvasViewTab *tab= [aTabView selectedPage];
  
  [tab makeVisible];
  
  [zoomPop selectItemAtIndex:[_canvas zoomLevel]];
}


static int toggles[]= {
  10, 11,
  20,
  30, 31, 32, 33, 34, 35,
  40, 41, 42
};

/*  
static NSString *toggleKeys[]= {
  @"esc",
  @"H",
  
  @"E",
  
  @"T",
  @"V",
  @"G",
  
  @"R",
  @"r",
  @"^R",
  
  @"I",
  @"N",
  @"L"
};
*/

- (BOOL)editCanvas:(MWEditCanvasView*)canvas keyDown:(NSEvent*)event
{
  int mods= [event modifierFlags] & (NSShiftKeyMask|NSControlKeyMask|NSCommandKeyMask|NSAlternateKeyMask);
  NSString *key= [event charactersIgnoringModifiers];

  if ([key isEqualTo:@"\e"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[0]] performClick:nil];
  }
  else if ([key isEqualTo:@"h"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[1]] performClick:nil];    
  }
  else if ([key isEqualTo:@"e"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[2]] performClick:nil];
  }
  else if ([key isEqualTo:@"t"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[3]] performClick:nil];
  }
  else if ([key isEqualTo:@"v"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[4]] performClick:nil];
  }
  else if ([key isEqualTo:@"g"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[5]] performClick:nil];
  }
  else if ([key isEqualTo:@"r"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[6]] performClick:nil];
  }    
  else if ([key isEqualTo:@"R"] && (mods & NSShiftKeyMask) == NSShiftKeyMask)
  {
    [[paletteView viewWithTag:toggles[7]] performClick:nil];
  }    
  else if ([key isEqualTo:@"r"] && (mods & NSControlKeyMask) == NSControlKeyMask)
  {
    [[paletteView viewWithTag:toggles[8]] performClick:nil];
  }    
  else if ([key isEqualTo:@"i"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[9]] performClick:nil];
  }    
  else if ([key isEqualTo:@"n"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[10]] performClick:nil];
  }    
  else if ([key isEqualTo:@"l"] && mods == 0)
  {
    [[paletteView viewWithTag:toggles[11]] performClick:nil];
  }   
  else if ([key isEqualTo:@" "] && (mods & NSShiftKeyMask) == NSShiftKeyMask)
  {
    [self toggleOverview:nil];
  }
  else
    return NO;
  return YES;
}  

/*
- (void)toggleToolHelp:(BOOL)show
{
  NSTextField *tip;
  NSFont *font= [NSFont systemFontOfSize:8.0];
  int i;
  
  if (show)
  {
    for (i= 0; i < sizeof(toggles)/sizeof(int); i++)
    {
      NSRect rect= [[paletteView viewWithTag:toggles[i]] frame];
      
      //rect.origin.y= 0;
      rect.origin.x= 1;
      rect.size.width= 30;
      rect.size.height= 10;
      
      tip= [[[NSTextField alloc] initWithFrame:rect] autorelease];
      [tip setDrawsBackground:NO];
      [tip setBordered:NO];
      [tip setAutoresizingMask:NSViewMinXMargin|NSViewMaxYMargin];
      [tip setFont:font];
      [tip setEditable:NO];
      [tip setStringValue:toggleKeys[i]];
      [tip setAlignment:NSLeftTextAlignment];
      [tip setTag:500+toggles[i]];
      [tip sizeToFit];
      [tip setAutoresizingMask:NSViewWidthSizable|NSViewMaxYMargin];
      [paletteView addSubview:tip];
    }
  }
  else
  {
    for (i= 0; i < sizeof(toggles)/sizeof(int); i++)
    {
      [[paletteView viewWithTag:500+toggles[i]] removeFromSuperview];
    }
  }
}
*/

- (IBAction)changeZoom:(id)sender
{
  [_canvas setZoomLevel:[sender indexOfSelectedItem]];
}


- (IBAction)pickImageFile:(id)sender
{
  NSOpenPanel *panel= [NSOpenPanel openPanel];
  
  [panel setTitle:@"Choose PNG Image File"];
  [panel setRequiredFileType:@"png"];
  if ([panel runModal] == NSFileHandlingPanelOKButton)
    [[imageOptions viewWithTag:1] setStringValue:[panel filename]];
}


- (void)updateToolOptions
{
  [_currentOptionsView removeFromSuperview];
  
  switch (_currentTool)
  {
    case MWToolCursor:
      _currentOptionsView= cursorOptions;
      break;
    case MWToolLayer:
      _currentOptionsView= layerOptions;
      break;
    case MWToolTable:
      _currentOptionsView= tableOptions;
      break;
    case MWToolImage:
      _currentOptionsView= imageOptions;
      break;
    case MWToolNote:
      _currentOptionsView= noteOptions;
      break;
    case MWToolView:
      _currentOptionsView= viewOptions;
      break;      
    case MWToolRoutine:
      _currentOptionsView= routineOptions;
      break;            
    default:
      _currentOptionsView= nil;
      break;
  }

  if (_currentOptionsView)
  {
    [[modelWindow contentView] addSubview:_currentOptionsView];
    [_currentOptionsView setFrameOrigin:NSMakePoint(285, NSHeight([[modelWindow contentView] frame]) - NSHeight([_currentOptionsView frame]))];
  }
}


- (IBAction)toggleGrid:(id)sender
{
  if ([sender state] == NSOnState)
    [_canvas setGridEnabled:YES];
  else
    [_canvas setGridEnabled:NO];
}


- (IBAction)alignObjects:(id)sender
{
  BOOL vert= NO;
  BOOL align= YES;
  NSString *side= nil;
  
  switch ([sender tag])
  {
    case 1:
      vert= YES;
      side= @"T";
      break;
    case 2:
      vert= YES;
      side= @"C";
      break;
    case 3:
      vert= YES;
      side= @"B";
      break;
    case 4:
      vert= NO;
      side= @"L";
      break;
    case 5:
      vert= NO;
      side= @"C";
      break;
    case 6:
      vert= NO;
      side= @"R";
      break;
      
    case 10:
      align= NO;
      vert= YES;
      break;
    case 11:
      align= NO;
      vert= NO;
      break;
  }
  
  if (align)
  {
    if (vert)
    {
      if (![_grt performModule:@"Workbench"
                     procedure:@"verticalAlign"
                     arguments:[NSArray arrayWithObject:side]])
        NSLog(@"error");
    }
    else
    {
      if (![_grt performModule:@"Workbench"
                     procedure:@"horizontalAlign"
                     arguments:[NSArray arrayWithObject:side]])
        NSLog(@"error");
    }
  }
  else
  {
    if (![_grt performModule:@"Workbench"
                   procedure:@"autoSpaceElements"
                   arguments:[NSArray arrayWithObjects:
                     [NSNumber numberWithInt:vert], 
                     [NSNumber numberWithInt:[[cursorOptions viewWithTag:20] intValue]], nil]])
      NSLog(@"error");
  }
}


- (IBAction)toggleTool:(id)sender
{
  int i;
  int tag= [sender tag];
  MWToolType oldTool;
  
  [[paletteView viewWithTag:toggles[_currentTool]] setState:NSOffState];
  
  oldTool= _currentTool;
  
  [sender setState:NSOnState];
  for (i= 0; i < sizeof(toggles)/sizeof(int); i++)
  {
    if (toggles[i] == tag)
    {
      [_canvas setCursor:[_toolCursors objectAtIndex:i]];
      _currentTool= (MWToolType)i;
      break;
    }
  }
  
  
  if (_relationshipContext)
  {
    [_grt performModule:@"Workbench"
              procedure:@"relationshipCancel"
              arguments:[NSArray arrayWithObject:[NSValue valueWithPointer:_relationshipContext]]];
    
    myx_grt_value_release(_relationshipContext);
    _relationshipContext= NULL;
  }
  
  switch (_currentTool)
  {
    case MWToolCursor:
      [_canvas setGrabPanning:NO];
      [statusText setStringValue:@"Cursor Tool"];
      break;
    case MWToolPanning:
      [_canvas setGrabPanning:YES];
      [statusText setStringValue:@"Panning Tool"];
      break;
    case MWToolTable:
      [_canvas setGrabPanning:NO];
      [statusText setStringValue:@"Table Tool: select position to place a new table."];
      break;
      
    case MWToolView:
      [_canvas setGrabPanning:NO];
      [statusText setStringValue:@"View Tool: select position to place a new view."];
      break;

    case MWToolRoutine:
      [_canvas setGrabPanning:NO];
      [statusText setStringValue:@"Routine Tool: select position to place new routine package."];
      break;      
      
    case MWToolRelationship1:
      
    case MWToolRelationship2:
    case MWToolRelationship3:
      _relationshipContext= [_grt performModule:@"Workbench"
                                       function:@"relationshipStart"
                                      arguments:[NSArray arrayWithObjects:
                           (_currentTool == MWToolRelationship1) ? @"11" : ((_currentTool == MWToolRelationship2) ? @"1n" : @"nm"),
                                        NSInt([[relationshipOptions viewWithTag:1] state] == NSOnState ? 1 : 0),
                                        NSInt([[relationshipOptions viewWithTag:2] state] == NSOnState ? 1: 0),
                                        nil
                                        ]];
      
      [statusText setStringValue:@"Relationship Tool: select the source table."];
      break;

    case MWToolLayer:
      [statusText setStringValue:@"Layer Tool: select an area for the new layer."];
      break;
  }
  [self updateToolOptions];
}


- (void)outlineView:(NSOutlineView *)outlineView willDisplayCell:(id)cell forTableColumn:(NSTableColumn *)tableColumn item:(id)item
{
  if (outlineView == layerTree)
  {
    if ([item isKindOfClass:[MWLayerTreeNode class]])
    {
//      [cell setImage:_];
    }
    else
    {
//      [cell setImage:];
    }
  }
}

- (BOOL)canvas:(MGenericCanvasView*)canvas handleMouseDrag:(NSPoint)pos
{
  return NO;
}


- (BOOL)canvas:(MGenericCanvasView*)canvas handleMouseDown:(int)button atPoint:(NSPoint)pos
{
  if (button == 0 && _currentTool == MWToolLayer)
  {
    [_canvas startAreaSelectionAt:pos];
  }
  else if (button == 0 && _currentTool == MWToolNote)
  {
    [_canvas startAreaSelectionAt:pos];
  }  

  if (_currentTool != MWToolCursor)
    return YES;
  else
    return NO;
}


- (BOOL)canvas:(MGenericCanvasView*)canvas handleMouseUp:(int)button atPoint:(NSPoint)pos
{
  MYX_GRT_VALUE *result;
  
  if (button == 0 && _currentTool == MWToolTable)
  {
    result= [_grt performModule:@"Workbench"
                       function:@"placeNewTableElement"
                      arguments:[NSArray arrayWithObjects:[NSNumber numberWithFloat:floor(pos.x)], 
                        [NSNumber numberWithFloat:floor(pos.y)],
                        [NSString stringWithFormat:@"table_%i",++_tableIndex],
                        [[[tableOptions viewWithTag:2] selectedItem] representedObject],
                        [[tableOptions viewWithTag:3] titleOfSelectedItem],
                        [[tableOptions viewWithTag:4] titleOfSelectedItem],
                        [[[tableOptions viewWithTag:1] selectedItem] representedObject],
                        nil]];
    if (result)
      myx_grt_value_release(result);
    
    [[paletteView viewWithTag:toggles[0]] performClick:nil];
  }
  else if (button == 0 && _currentTool == MWToolView)
  {
    result= [_grt performModule:@"Workbench"
                       function:@"placeNewViewElement"
                      arguments:[NSArray arrayWithObjects:[NSNumber numberWithFloat:floor(pos.x)], 
                        [NSNumber numberWithFloat:floor(pos.y)],
                        [NSString stringWithFormat:@"view_%i",++_viewIndex],
                        [[[viewOptions viewWithTag:2] selectedItem] representedObject],
                        [[[viewOptions viewWithTag:1] selectedItem] representedObject],
                        nil]];
    if (result)
      myx_grt_value_release(result);
    
    [[paletteView viewWithTag:toggles[0]] performClick:nil];
  }  
  else if (button == 0 && _currentTool == MWToolRoutine)
  {
    result= [_grt performModule:@"Workbench"
                       function:@"placeNewRoutineGroupElement"
                      arguments:[NSArray arrayWithObjects:[NSNumber numberWithFloat:floor(pos.x)], 
                        [NSNumber numberWithFloat:floor(pos.y)],
                        [NSString stringWithFormat:@"routineGroup_%i",++_routineIndex],
                        [[[routineOptions viewWithTag:2] selectedItem] representedObject],
//                        [[[routineOptions viewWithTag:1] selectedItem] representedObject],
                        nil]];
    if (result)
      myx_grt_value_release(result);
    
    [[paletteView viewWithTag:toggles[0]] performClick:nil];
  }    
  else if (button == 0 && _currentTool == MWToolLayer)
  {
    NSRect rect= [_canvas finishAreaSelection];
    result= [_grt performModule:@"Workbench"
                       function:@"placeNewLayerElement"
                      arguments:[NSArray arrayWithObjects:
                        [NSNumber numberWithFloat:floor(rect.origin.x)],
                        [NSNumber numberWithFloat:floor(rect.origin.y)],
                        [NSNumber numberWithFloat:floor(rect.size.width)], 
                        [NSNumber numberWithFloat:floor(rect.size.height)],
                        [NSString stringWithFormat:@"Layer %i", ++_layerIndex],
                        [[[layerOptions viewWithTag:1] selectedItem] representedObject],
                        nil]];
    if (result)
      myx_grt_value_release(result);
    
    [[paletteView viewWithTag:toggles[0]] performClick:nil];
  }
  else if (button == 0 && _currentTool == MWToolNote)
  {
    NSRect rect= [_canvas finishAreaSelection];
    result= [_grt performModule:@"Workbench"
                       function:@"placeNewNoteElement"
                      arguments:[NSArray arrayWithObjects:
                        [NSNumber numberWithFloat:floor(rect.origin.x)], 
                        [NSNumber numberWithFloat:floor(rect.origin.y)],
                        [NSNumber numberWithFloat:floor(rect.size.width)], 
                        [NSNumber numberWithFloat:floor(rect.size.height)],
                        [[[noteOptions viewWithTag:1] selectedItem] representedObject],
                        nil]];
    if (result)
      myx_grt_value_release(result);
    
    [[paletteView viewWithTag:toggles[0]] performClick:nil];
  }
  else if (button == 0 && _currentTool == MWToolEraser)
  {
    MYX_GRT_VALUE *result;
    result= [_grt performModule:@"Workbench"
                       function:@"deleteObjectAtPoint"
                      arguments:[NSArray arrayWithObjects:[NSNumber numberWithFloat:floor(pos.x)], 
                        [NSNumber numberWithFloat:floor(pos.y)],
                        nil]];
    if (result)
    {
      myx_grt_value_release(result);
      
      [[paletteView viewWithTag:toggles[0]] performClick:nil];
    }
  }    
  else if (button == 0 && (_currentTool == MWToolRelationship1
                           || _currentTool == MWToolRelationship2
                           || _currentTool == MWToolRelationship3))
  {
    NSString *message;
    
    message= [_grt performModule:@"Workbench"
                  stringFunction:@"relationshipClickedPoint"
                       arguments:[NSArray arrayWithObjects:
                         [NSValue valueWithPointer:_relationshipContext],
                         [NSNumber numberWithFloat:floor(pos.x)],
                         [NSNumber numberWithFloat:floor(pos.y)],
                         nil]];
    
    if (message)
      [statusText setStringValue: message];
    
    if (myx_grt_dict_item_get_as_int(_relationshipContext, "step") == -1)
    {
      [[paletteView viewWithTag:toggles[0]] performClick:nil];
    }
  }
  
  // handle any eventual layer/element relocations
  if (_currentTool == MWToolCursor)
    myx_grt_wb_bridge_process_relocations([_grt grt]);
    
  return NO;
}


- (void)objectEditorSaved:(MGRTObjectEditor*)editor
{        
  if ([editor isKindOfClass:[MGRTTableEditor class]])
  {
    [_grt performModule:@"Workbench"
               function:@"createRelationshipsFromFKs"
              arguments:[NSArray arrayWithObject:[NSValue valueWithPointer:[_grt globalValue:"/workbench/model/views/0"]]]];
  }
}


- (void)objectEditorClosed:(MGRTObjectEditor*)editor
{
  [_openEditors removeObjectForKey:[NSString stringWithUTF8String:myx_grt_dict_id_item_as_string([editor editedObject])]];
}


@end
